/*
 *  Ipd version 1.0, Copyright (C) 1995 G. Hutchings
 *  Ipd comes with ABSOLUTELY NO WARRANTY.
 *  This is free software, and you are welcome to redistribute it
 *  under certain conditions; see the file COPYING for details.
 */

#ifndef lint
static char rcsid[] = "$Id: rule.c,v 1.2 1995/06/08 16:21:38 glen Exp $";
#endif

/* Rule functions */

#include <stdio.h>
#include "config.h"
#include "rule.h"
#include "proto.h"

/* The list of rules available */

struct rule Rules[] = {
    { "All C",				all_c,		'C' },
    { "All D",				all_d,		'D' },
    { "Downing",			downing,	'd' },
    { "Forgiving Tit for Tat",		ftft,		'%' },
    { "Friedman",			friedman,	'f' },
    { "Joss",				joss,		'j' },
    { "Random",				rnd,		'R' },
    { "Revised Downing",		r_downing,	'r' },
    { "Revised State Transition",	rst,		's' },
    { "Suspicious Tit for Tat",		stft,		',' },
    { "Tester",				tester,		't' },
    { "Tit for Tat",			tft,		'.' },
    { "Tit for Two Tats",		tf2t,		':' },
    { "Tranquilizer",			tranq,		'q' },
    { "Two Tits for Tat",		t2ft,		';' },
    { NULL,				NULL,		' ' },
};

/* The rules */

/*
 * Random
 *
 * The benchmark `bad' rule.
 */
int
rnd(int turn, int *me, int *you)
{
    return (RANDOM(2) == 0 ? COOPERATE : DEFECT);
}

/*
 * All C
 *
 * Always co-operate regardless.  About as nice as it's possible
 * to get.
 */
int
all_c(int turn, int *me, int *you)
{
    return COOPERATE;
}

/*
 * All D
 *
 * Always defect regardless.  The ultimate nasty rule.
 */
int
all_d(int turn, int *me, int *you)
{
    return DEFECT;
}

/*
 * Tit for tat
 *
 * The champion of Axelrod's contests.  It co-operates on the first
 * move, and then does whatever the other player did on the previous
 * move.
 */
int
tft(int turn, int *me, int *you)
{
    return (turn == 0 ? COOPERATE : you[turn - 1]);
}

/*
 * Suspicious tit for tat
 *
 * The same as tit-for-tat, except that it defects on the first move.
 */
int
stft(int turn, int *me, int *you)
{
    return (turn == 0 ? DEFECT : you[turn - 1]);
}

/*
 * Tit for two tats
 *
 * More forgiving than tit-for-tat, it will allow two consecutive
 * defections before defecting.
 */
int
tf2t(int turn, int *me, int *you)
{
    if (turn < 2)
	return COOPERATE;
    if (you[turn - 1] == COOPERATE)
	return COOPERATE;
    if (you[turn - 2] == COOPERATE)
	return COOPERATE;
    return DEFECT;
}

/*
 * Two tits for tat
 *
 * Less forgiving than tit-for-tat.  It will reply to a defection with
 * two consecutive defections.
 */
int
t2ft(int turn, int *me, int *you)
{
    if (turn == 0)
	return COOPERATE;
    if (turn == 1)
	return you[turn - 1];
    if (you[turn - 1] == DEFECT)
	return DEFECT;
    if (you[turn - 2] == DEFECT)
	return DEFECT;
    return COOPERATE;
}

/*
 * Forgiving tit for tat
 *
 * Slightly more forgiving than tit-for-tat.  It will defect following
 * a defection with probability only 9/10.
 */
int
ftft(int turn, int *me, int *you)
{
    if (turn == 0)
	return COOPERATE;
    if (you[turn - 1] == COOPERATE)
	return COOPERATE;
    return (RANDOM(10) == 0 ? COOPERATE : DEFECT);
}

/*
 * Joss
 *
 * Plays virtually the same as tit-for-tat, except that occasionally
 * it will try an unprovoked defection.
 */
int
joss(int turn, int *me, int *you)
{
    if (turn == 0)
	return COOPERATE;
    if (you[turn - 1] == COOPERATE)
	return (RANDOM(10) == 0 ? DEFECT : COOPERATE);
    return DEFECT;
}

/*
 * Friedman
 *
 * This will co-operate until it receives a defection, and always
 * defect from then on - completely unforgiving.
 */
int
friedman(int turn, int *me, int *you)
{
    static int defected;

    if (turn == 0) {
	defected = 0;
	return COOPERATE;
    } else if (you[turn - 1] == DEFECT) {
	defected = 1;
    }

    return (defected ? DEFECT : COOPERATE);
}

/*
 * Tester
 *
 * This will see what it can get away with on the first turn.  If
 * the opponent ever defects, it will apologise and play tit-for-tat
 * for the rest of the game.
 */
int
tester(int turn, int *me, int *you)
{
    static int defected;

    if (turn == 0) {
	defected = 0;
	return DEFECT;
    } else if (!defected && you[turn - 1] == DEFECT) {
	defected = 1;
	return COOPERATE;
    }

    if (defected)
	return you[turn - 1];

    return (turn < 3 ? COOPERATE : DEFECT);
}

/*
 * Downing
 *
 * This rule tries to be clever and estimate the probabilities of
 * co-operation by the opponent after it co-operates or defects.
 * If the opponent appears to be unresponsive (both probabilities
 * are nearly the same), it always defects.  Otherwise it co-operates.
 */
int
downing(int turn, int *me, int *you)
{
    static double cc, cd;
    static int nc, nd;
    double pc, pd;

    if (turn < 2) {
	nc = nd = 1;
	cc = cd = 0.5;
    } else if (me[turn - 2] == COOPERATE) {
	nc++;
	if (you[turn - 1] == COOPERATE) cc++;
    } else {
	nd++;
	if (you[turn - 1] == COOPERATE) cd++;
    }

    pc = cc / nc;
    pd = cd / nd;

    return (pc > pd + 0.2 ? COOPERATE : DEFECT);
}

/*
 * Revised Downing
 *
 * This is a modification of Downing that tries to fix a problem
 * with it: it initially assumes the opponent will be unresponsive.
 */
int
r_downing(int turn, int *me, int *you)
{
    static double cc, cd;
    static int nc, nd;
    double pc, pd;

    if (turn < 2) {
	nc = nd = 5;
	cc = 4.0;
	cd = 1.0;
    } else if (me[turn - 2] == COOPERATE) {
	nc++;
	if (you[turn - 1] == COOPERATE) cc++;
    } else {
	nd++;
	if (you[turn - 1] == COOPERATE) cd++;
    }

    pc = cc / nc;
    pd = cd / nd;

    return (pc > pd + 0.2 ? COOPERATE : DEFECT);
}

/*
 * Tranquilizer
 *
 * This plays tit-for-tat for the first few moves, and then
 * tries to get away with something.  If they don't respond,
 * increase probability of defection - but never do two
 * unprovoked defections in a row.
 */
int
tranq(int turn, int *me, int *you)
{
    static int defect_prob, just_defected;

    /* Initialise */
    if (turn == 0) {
	defect_prob = 10;
	just_defected = 0;
    }

    /* Play tit-for-tat for the first few moves */
    if (turn < 10)
	return (turn == 0 ? COOPERATE : you[turn - 1]);

    /* Defect if provoked */
    if (you[turn - 1] == DEFECT) {
	just_defected = 0;
	return DEFECT;
    }

    /* Now do occasional unprovoked defections */
    if (!just_defected && RANDOM(defect_prob)) {
	just_defected = 1;
	return DEFECT;
    } else {
	/* They co-operated after a defection - do more! */
	if (defect_prob > 4)
	    defect_prob -= 2;
	just_defected = 0;
    }

    return COOPERATE;
}

/*
 * Revised State Transition
 *
 * This plays tit-for-tat until it has enough information to
 * decide what's probably best.
 */
int
rst(int turn, int *me, int *you)
{
    static double cc, pcc, cd, pcd, dc, pdc, dd, pdd;
    static int ncc, ncd, ndc, ndd;

    /* Initialise */
    if (turn == 0) {
	cc = ncc = 0;
	cd = ncd = 0;
	dc = ndc = 0;
	dd = ndd = 0;
	return COOPERATE;
    }

    /* Observe responses */
    if (turn > 2) {
	if (you[turn - 2] == COOPERATE && me[turn - 2] == COOPERATE) {
	    ncc++;
	    if (you[turn - 1] == COOPERATE) cc++;
	    pcc = cc / ncc;
	} else if (you[turn - 2] == COOPERATE && me[turn - 2] == DEFECT) {
	    ncd++;
	    if (you[turn - 1] == COOPERATE) cd++;
	    pcd = cd / ncd;
	} else if (you[turn - 2] == DEFECT && me[turn - 2] == COOPERATE) {
	    ndc++;
	    if (you[turn - 1] == COOPERATE) dc++;
	    pdc = dc / ndc;
	} else {
	    ndd++;
	    if (you[turn - 1] == COOPERATE) dd++;
	    pdd = dd / ndd;
	}
    }

    /* Play tit-for-tat until enough info obtained */
    if (turn < 10)
	return you[turn - 1];

    /* If you've never defected, assume the best */
    if (ndc == 0 && ndd == 0)
	return COOPERATE;

    /* If you tend to co-operate after I do, then co-operate */
    if (ncc > 0 && pcc > 0.7 && ndc > 0 && pdc > 0.7)
	return COOPERATE;

    /* If you tend to co-operate after I defect, do tit-for-tat */
    if (ncd > 0 && pcd > 0.7 && ndd > 0 && pdd > 0.7)
	return you[turn - 1];

    /* You are a cad, sir */
    return DEFECT;
}
